/* $Header: /ext/yav/catty/xpx/RCS/convert.c,v 1.11 1994/01/27 14:48:17 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#define col3to8(x) (((x)<<5)|((x)<<2)|((x)>>1))

/* return : read color set */
read_ccb(fp, p)
     FILE *fp; COL *p;
{
  unsigned char buf[2*MAXCOLSET];
  int i, r;

  r = 0;
  while (fread(buf, 2*MAXCOLSET, 1, fp) != 0) {
    for (i = 0; i < MAXCOLSET; i++) {
      p->blue = buf[i*2] & 7;
      p->red = (buf[i*2]>>3) & 7;
      p->green = ((buf[i*2]>>6) & 3)|((buf[i*2+1]&1) << 2);
      p->blue = col3to8(p->blue);
      p->red = col3to8(p->red);
      p->green = col3to8(p->green);
      p++;
    }
    r++;
  }
  return r;
}

/* return pnm type number */
read_pnm_header(fp, data, colorhints, nhints)
     FILE *fp;
     int *data;
     COL *colorhints;
     int *nhints;
{
  int i, result, r, g, b;
  int datan, maxcol;
  char buf[256], *p;

  if (fgets(buf, sizeof(buf), fp) == NULL)
    return 0;
  if (buf[0] != 'P') {
    fprintf(stderr, "not ppm file!\n");
    return 0;
  }
  switch (buf[1]) {
  case '2':
    result = 2;
    break;
  case '3':
    result = 3;
    break;
  case '5':
    result = 5;
    break;
  case '6':
    result = 6;
    break;
  default:
    fprintf(stderr, "not supported ppm,\n");
    return 0;
  }
  maxcol = -1;
  datan = 0;
  while (datan < 3 && fgets(buf, sizeof(buf), fp) != NULL) {
    if (buf[0] == '#') {
      if (colorhints != NULL && nhints != NULL) {
	if (sscanf(buf, "# color:%d %d %d %d", &i, &r, &g, &b) == 4) {
	  if (i < *nhints) {
	    (colorhints+i)->red = r;
	    (colorhints+i)->green = g;
	    (colorhints+i)->blue = b;
	    if (maxcol < i)
	      maxcol = i;
	  }
	}
      }
    } else {
      for (p = buf; (i = strtol(p, &p, 10)) != 0 && datan < 3; datan++)
	data[datan] = i;
    }
  }
  if (colorhints != NULL && nhints != NULL)
    *nhints = maxcol+1;
  return result;
}

read_pnm_pixels(fp, t, cp, n)
     FILE *fp;
     int t;
     COL *cp;
     int n;
{
  int i, x, rgbsw;
  char *p, *op;
  char buf[256];

  rgbsw = 0;
  switch(t) {
  case 3:
    p = buf;
    *p = '\0';
    for (x = 0; x < n; x++) {
      if (*p == '\0' || *p == '\n') {
	if (fgets(buf, sizeof(buf), fp) == NULL)
	  return x;
	p = buf;
	continue;
      }
      op = p;
      i = strtol(p, &p, 10);
      if (p == op)
	continue;
      switch(rgbsw) {
      case 0:
	cp->red = i;
	rgbsw = 1;
	break;
      case 1:
	cp->green = i;
	rgbsw = 2;
	break;
      case 2:
	cp->blue = i;
	cp++;
	rgbsw = 0;
	break;
      }
    }
    break;
  case 6:
    return fread(cp, 3, n, fp);
  default:
    fprintf(stderr, "pnm type %d!\n", t);
    return 0;
  }
  /* NOT REACHED */
}

static COL lastcol;
static int lastcol_pixel;

color_pixel(tbl, col)
     COL *tbl;
     COL *col;
{
 int i;
 
 if (lastcol_pixel >= 0 &&
     col->red == lastcol.red &&
     col->green == lastcol.green &&
     col->blue == lastcol.blue)
   return lastcol_pixel;
 for (i = 0; i < MAXCOLSET*MAXPALSET; i++) {
   if ((coltbl[i] & (CM_ALLOCED|CM_REQUEST)) &&
       col->red == tbl->red &&
       col->green == tbl->green &&
       col->blue == tbl->blue) {
     lastcol = *col;
     return lastcol_pixel = i;
   }
   tbl++;
 }
 return -1;
}

read_pnm(fp)
     FILE *fp;
{
  int c, i, n, f, x, y, pnmtype;
  int data[3], w, h;
  int colcnt;
  COL px[640];
  
  colcnt = (MAXPALSET-current_pal)*16;
  pnmtype = read_pnm_header(fp, data, color_buf+16*current_pal, &colcnt);
  if (pnmtype == 0)
    return 1;
  free_color();
  /* magtoppm commented color information set */
  get_color(color_buf+16*current_pal, colcnt);

  w = data[0];
  h = data[1];
  if (w > imgmaxw) {
    w = imgmaxw;
    fprintf(stderr, "width %d trancate to %d.\n", data[0], w);
  }
  if (h > imgmaxh) {
    h = imgmaxh;
    fprintf(stderr, "height %d trancate to %d.\n", data[1], h);
  }
  set_image_size(w, h);
  lastcol_pixel = -1;
  for (y = 0; y < h; y++) {
    f = 1;
    for (n = data[0]; n; n -= i) {
      i = n > w ? w : n;
      read_pnm_pixels(fp, pnmtype, px, i);
      if (f) {
	for (x = 0; x < i; x++) {
	  c = color_pixel(color_buf+16*current_pal, &px[x]);
	  if (c < 0)
	    c = request_color(&px[x]);
	  *(imgdata+y*imgmaxw+x) = c;
	}
	f = 0;
      }
    }
  }
  get_requested_colors();
  image_update_all();
  return 0;
}

static char kiss_magic_number[] = "KiSS";

read_kcf(fp)
     FILE *fp;
{
  int i, s, bpc, cc, col, set, preload;
  COL *cp;
  unsigned char *p, buf[4096];

  /* old style kcf spec set */
  bpc = 12;
  col = 16;
  set = 10;
  preload = strlen(kiss_magic_number);
  if (!fread(buf, preload, 1, fp))
    return 1;
  if (strncmp(buf, kiss_magic_number, preload)==0) {
    /* read new style kcf header section */
    if (!fread(buf+preload, 32-preload, 1, fp))
      return 1;
    /* check color file mark */
    if (buf[4] != 0x10)
      return 1;
    bpc = buf[5];
    col = buf[8]|(buf[9]<<8);
    set = buf[10]|(buf[11]<<8);
    preload = 0;
  }
  cp = color_buf;
  cc = ((bpc+7)>>3)*col;
  for (s = 0; s < set; s++) {
    if (!fread(buf+preload, cc-preload, 1, fp))
      return 1;
    p = buf;
    for (i = 0; i < col; i++) {
      if (bpc == 12) {
	cp->red = (*p&0xf0)|(*p>>4);
	cp->blue = (*p&0x0f)|((*p&0x0f)<<4);
	p++;
	cp->green = (*p&0x0f)|((*p&0x0f)<<4);
	p++;
      } else {
	cp->red = *p++;
	cp->green = *p++;
	cp->blue = *p++;
      }
      cp++;
    }
    preload = 0;
  }
  free_color();
  get_color(color_buf+16*current_pal, col);
  image_update_col();
  return 0;
}

write_kcf(fp, newver)
     FILE *fp;
{
  int i, s, bpc, col, set, cc;
  COL *cp;
  unsigned char *p, buf[4096];

  /* old style kcf spec set */
  bpc = 12;
  col = 16;
  set = 10;
  if (newver) {
    bpc = 24;
    memset(buf, 0, 32);
    memcpy(buf, kiss_magic_number, strlen(kiss_magic_number));
    /* set color file mark */
    buf[4] = 0x10;
    buf[5] = bpc;
    buf[8] = col;
    buf[9] = col>>8;
    buf[10] = set;
    buf[11] = set>>8;
    if (!fwrite(buf, 32, 1, fp))
      return 1;
  }
  cp = color_buf;
  cc = ((bpc+7)>>3)*col;
  for (s = 0; s < set; s++) {
    p = buf;
    for (i = 0; i < col; i++) {
      if (bpc == 12) {
	*p++ = (cp->red & 0xf0)|(cp->blue >> 4);
	*p++ = cp->green >> 4;
      } else {
	*p++ = cp->red;
	*p++ = cp->green;
	*p++ = cp->blue;
      }
      cp++;
    }
    if (!fwrite(buf, cc, 1, fp))
      return 1;
  }
  return 0;
}

read_cel(fp)
     FILE *fp;
{
  int c, x, y, w, h, preload, bpp;
  int dataw, datah;
  unsigned char buf[4096];

  preload = strlen(kiss_magic_number);
  if (!fread(buf, preload, 1, fp))
    return 1;
  if (strncmp(buf, kiss_magic_number, preload) == 0) {
    /* read new style cel file header section */
    if (!fread(buf+preload, 32-preload, 1, fp))
      return 1;
    /* check cel file mark */
    if (buf[4] != 0x20)
      return 1;
    bpp = buf[5];
    w = buf[8]|(buf[9]<<8);
    h = buf[10]|(buf[11]<<8);
  } else {
    bpp = 4;
    w = buf[0]|(buf[1]<<8);
    h = buf[2]|(buf[3]<<8);
  }
  dataw = w;
  datah = h;
  if (w > imgmaxw) {
    w = imgmaxw;
    fprintf(stderr, "width %d trancate to %d.\n", dataw, w);
  }
  if (h > imgmaxh) {
    h = imgmaxh;
    fprintf(stderr, "height %d trancate to %d.\n", datah, h);
  }
  set_image_size(w, h);
  for (y = 0; y < h; y++) {
    if (!fread(buf, bpp==4?((dataw+1)>>1):dataw, 1, fp))
      return 1;
    for (x = 0; x < w; x++) {
      if (bpp == 4) {
	c = buf[x>>1];
	if ((x & 1) == 0)
	  c >>= 4;
	else
	  c &= 0x0f;
      } else {
	c = buf[x];
      }
      *(imgdata+y*imgmaxw+x) = c;
    }
  }
  image_update_all();
  return 0;
}

write_cel(fp, newver)
     FILE *fp;
{
  int c, x, y, w, h, bpp;
  unsigned char buf[4096];

  w = imgfilew;
  h = imgfileh;
  bpp = 4;
  if (newver) {
    memset(buf, 0, 32);
    memcpy(buf, kiss_magic_number, strlen(kiss_magic_number));
    /* set cel file mark */
    buf[4] = 0x20;
    buf[5] = bpp;
    buf[8] = w;
    buf[9] = w >> 8;
    buf[10] = h;
    buf[11] = h >> 8;
    if (!fwrite(buf, 32, 1, fp))
      return 1;
  } else {
    buf[0] = w;
    buf[1] = w >> 8;
    buf[2] = h;
    buf[3] = h >> 8;
    if (!fwrite(buf, 4, 1, fp))
      return 1;
  }
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x++) {
      c = *(imgdata+y*imgmaxw+x);
      if (bpp == 4) {
	if ((x & 1) == 0)
	  buf[x>>1] = c<<4;
	else
	  buf[x>>1] |= c;
      } else {
	buf[x] = c;
      }
    }
    if (!fwrite(buf, bpp==4?((w+1)>>1):w, 1, fp))
      return 1;
  }
  return 0;
}

write_xbm_out(fp, data, datalen, end)
     FILE *fp;
     unsigned char *data;
{
  int i;
  char *p;
  char buf[256];

  strcpy(buf, "  ");
  p = buf+2;
  for (i = 0; i < datalen; i++) {
    sprintf(p, " 0x%02x,", *data++);
    p += strlen(p);
  }
  if (end)
    strcpy(p-1, "};");
  strcat(buf, "\n");
  return fputs(buf, fp);
}

/* write X11 bitmap */
write_xbm(fp, name)
     FILE *fp;
     char *name;
{
  int i, x, y, w, h;
  int outcount;
  unsigned char c;
  unsigned char outbuf[12];
  static unsigned char bittbl[8] = {0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};

  w = imgfilew;
  h = imgfileh;
  fprintf(fp, "#define %s_width %d\n", name, w);
  fprintf(fp, "#define %s_height %d\n", name, h);
  fprintf(fp, "static char %s_bits[] = {\n", name);
  outcount = 0;
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x += 8) {
      c = 0;
      for (i = 0; i < 8; i++) {
	if ((x + i < w)&& *(imgdata+y*imgmaxw+x+i))
	  c |= bittbl[i];
      }
      if (outcount >= sizeof(outbuf)) {
	write_xbm_out(fp, outbuf, outcount, 0);
	outcount = 0;
      }
      outbuf[outcount++] = c;
    }
  }
  write_xbm_out(fp, outbuf, outcount, 1);
  return 0;
}

/* read X11 bitmap */
read_xbm(fp)
     FILE *fp;
{
  int i, x, y, w, h, c;
  static unsigned char bittbl[8] = {0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};
  char buf[1024], *p;
  char namebuf[256];

  w = h = 0;
  while (fgets(buf, sizeof(buf), fp) != NULL) {
    if (strncmp(buf, "static", 6) == 0)
      break;
    if (sscanf(buf, "#define %s %d", namebuf, &c) == 2) {
      i = strlen(namebuf);
      if (i >= 5 && strcmp(namebuf+i-5, "width") == 0)
	w = c;
      else if (i >= 6 && strcmp(namebuf+i-6, "height") == 0)
	h = c;
    }
  }
  if (!w || !h)
    return 1;
  y = x = 0;
  p = buf;
  *p = '\0';
  for (;;) {
    while (*p && (*p == ' ' || *p == '\t' || *p == ','))
      p++;
    if (*p == '\0' || *p == '\n') {
      if (fgets(buf, sizeof(buf), fp) == NULL)
	break;
      p = buf;
    }
    if (sscanf(p, "0x%02x", &c)) {
      p += 4;
      for (i = 0; i < 8; i++) {
	if (x+i < w)
	  *(imgdata+y*imgmaxw+x+i) = ((c & bittbl[i]) ? 1 : 0);
      }
      x += 8;
      if (x >= w) {
	if (++y >= h)
	  break;
	x = 0;
      }
    } else {
      p++;
    }
  }
  set_image_size(w, h);
  image_update_all();
  return 0;
}

gen_fontpix(p, cbb)
     unsigned char *p; unsigned char *cbb;
{
  int x, y;
  static unsigned char bittbl[8] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

  for (y = 0; y < 8; y++) {
    for (x = 0; x < 8; x++) {
      if (*cbb & bittbl[x])
	*p = 1;
      else
	*p = 0;
      if (*(cbb+1) & bittbl[x])
	*p |= 2;
      if (*(cbb+16) & bittbl[x])
	*p |= 4;
      if (*(cbb+17) & bittbl[x])
	*p |= 8;
      p++;
    }
    cbb += 2;
  }
}

gen_image_cbb(cbb)
     unsigned char *cbb;
{
  int dx, dy, xx, yy, ix, iy, x, y;
  unsigned char fontpix[8*8], *p;
  int c;

  for (y = 0; y < 8*16; y+=8) {
    for (x = 0; x < 8*16; x+=8) {
      gen_fontpix(fontpix, cbb);
      cbb += 32;
      p = fontpix;
      for (iy = 0; iy < 8; iy++) {
	for (ix = 0; ix < 8; ix++) {
	  *(imgdata+(y+iy)*imgmaxw+x+ix) = *p++;
	}
      }
    }
  }
}

/* set XImage from imgdata */
setpixel_sub2(x0, y0, pixel)
{
  int x, y, x1, y1;

  x0 *= zoomfactor;
  y0 *= zoomfactor;
  x1 = x0+zoomfactor;
  y1 = y0+zoomfactor;
  for (y = y0; y < y1; y++)
    for (x = x0; x < x1; x++)
      XPutPixel(img, x, y, pixel);
}
  
/* set XImage from imgdata */
setpixel_sub(x0, y0, pixel)
{
  int x, y, x1, y1;

  if (vimg != NULL)
    XPutPixel(vimg, x0, y0,  pixel);
  x0 *= zoomfactor;
  y0 *= zoomfactor;
  x1 = x0+zoomfactor;
  y1 = y0+zoomfactor;
  for (y = y0; y < y1; y++)
    for (x = x0; x < x1; x++)
      XPutPixel(img, x, y, pixel);
}
  
/* set view image */
#define setpixel_subv(x0,y0,px) XPutPixel(vimg,(x0),(y0),(px))

setpixelcolor(x, y, col)
{
  *(imgdata+y*imgmaxw+x) = col;
  setpixel_sub2(x, y, pixtbl[col]);
  if (vimg != NULL)
    setpixel_subv(x, y, pixtbl[col]);
}

gen_image_subv(x0, y0, w, h)
{
  int x, y;
  unsigned char *ip;
  
  for (y = y0; y < y0+h; y++) {
    ip = imgdata+y*imgmaxw+x0;
    for (x = x0; x < x0+w; x++) {
      setpixel_subv(x, y, pixtbl[*ip++]);
    }
  }
}

/* set XImage from imgdata */
gen_image_sub(x0, y0, w, h)
{
  int x, y;
  unsigned char *ip;
  
  for (y = y0; y < y0+h; y++) {
    ip = imgdata+y*imgmaxw+x0;
    for (x = x0; x < x0+w; x++) {
      setpixel_sub2(x, y, pixtbl[*ip++]);
    }
  }
}

copy_area(dstx, dsty, srcx, srcy, w, h)
{
  int x, y;
  
  if (dsty*imgmaxw+dstx > srcy*imgmaxw+srcx) {
    for (y = h-1; y >= 0; y--) {
      for (x = w-1; x >= 0; x--) {
	*(imgdata+(dsty+y)*imgmaxw+dstx+x) =
	  *(imgdata+(srcy+y)*imgmaxw+srcx+x);
      }
    }
  } else {
    for (y = 0; y < h; y++) {
      for (x = 0; x < w; x++) {
	*(imgdata+(dsty+y)*imgmaxw+dstx+x) =
	  *(imgdata+(srcy+y)*imgmaxw+srcx+x);
      }
    }
  }
  gen_image_sub(dstx, dsty, w, h);
  if (vimg != NULL)
    gen_image_subv(dstx, dsty, w, h);
}
  

gen_image_col(x0, y0, w, h)
{
  int x, y;
  unsigned char *ip;
  
  for (y = y0; y < y0+h; y++) {
    ip = imgdata+y*imgmaxw+x0;
    for (x = x0; x < x0+w; x++) {
      if (coltbl[*ip] & CM_UPDATING) {
	coltbl[*ip] |= CM_UPDATED;
	setpixel_sub2(x, y, pixtbl[*ip++]);
      }
    }
  }
  if (vimg != NULL) {
    for (y = y0; y < y0+h; y++) {
      ip = imgdata+y*imgmaxw+x0;
      for (x = x0; x < x0+w; x++) {
	if (coltbl[*ip] & CM_UPDATING) {
	  coltbl[*ip] |= CM_UPDATED;
	  setpixel_subv(x, y, pixtbl[*ip++]);
	}
      }
    }
  }
  for (x = 0; x < MAXPALSET*MAXCOLSET; x++) {
    if (coltbl[x] & CM_UPDATED)
      coltbl[x] &= ~(CM_UPDATED|CM_UPDATING);
  }
}
  
/* End of file */
