/* $Header: /ext/yav/catty/xpx/RCS/edit.c,v 1.2 1994/02/02 07:16:04 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#include "editdef.h"
#include "menudef.h"

#define BDW 1

#define EDITWINW 128
#define EDITWINH 128

#define EDMW	64
#define EDMH	24

static MENU menutbl[] = {
  {
    EDIT_PLOT,
    4, 4+(EDMH+BDW)*0, EDMW, EDMH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, MenuModeMaskOn, BDW,
    NULL, "Plot"},
  {
    EDIT_COPY,
    4, 4+(EDMH+BDW)*1, EDMW, EDMH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "Copy"},
  {
    EDIT_OK,
    EDITWINW-(EDMW+BDW*2)-4, EDITWINH-(EDMH+BDW*2)-4, EDMW, EDMH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, SouthEastGravity,
    0, 0, BDW,
    NULL, "OK"},
  {
    -1,
    0, 0, 0, 0,
    0, 0, 0,
    0, 0, 0,
    NULL, ""}};

static Window editwin;
static int editwin_opened = 0;

create_edit_window(x, y)
{
  int i;
  XSizeHints hint;
  char *p;
  char buf[256];

  if (editwin_opened) {
    XMapWindow(dsp, editwin);
    return 1;
  }
  hint.x = x; hint.y = y;
  hint.width = EDITWINW; hint.height = EDITWINH;
  hint.flags = PPosition|PSize;
  editwin = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
				hint.width, hint.height, BDW, fg, bg);
  sprintf(buf, "%s edit", myname);
  XSetStandardProperties(dsp, editwin, buf, buf, None, NULL, NULL, &hint);
  XSelectInput(dsp, editwin, ExposureMask);
  XMapRaised(dsp, editwin);

  menuwin_create(menutbl, editwin, 0, 0);
  editwin_opened = 1;
  return 0;
}

unmap_edit_window()
{
  XUnmapWindow(dsp, editwin);
}

close_edit_window()
{
  if (!editwin_opened)
    return;
  editwin_opened = 0;
  menuwin_destroy(menutbl);
  XDestroyWindow(dsp, editwin);
}

edit_expose(ev)
     XEvent *ev;
{
  int i;
  char buf[256];

  if (!editwin_opened)
    return;
  menuwin_expose(menutbl, ev);
}

edit_mode_change(n)
{
  if (!editwin_opened)
    return;
  if (n != edit_mode) {
    menutbl[edit_mode].mode &= ~MenuModeMaskOn;
    XClearArea(dsp, menutbl[edit_mode].win, 0, 0, 0, 0, True);
    edit_mode = n;
  }
  menutbl[n].mode |= MenuModeMaskOn;
  XClearArea(dsp, menutbl[n].win, 0, 0, 0, 0, True);
}

edit_press(ev)
     XEvent *ev;
{
  int i, oldsel;

  if (!editwin_opened)
    return;
  i = menuwin_press(menutbl, ev);
  if (i >= 0) {
    if (i == EDIT_OK) {
      menutbl[i].mode &= ~MenuModeMaskOn;
      menu_stat_change(MENU_EDIT, 0);
    } else {
      edit_mode_change(i);
    }
  }
}

edit_enter(ev)
     XEvent *ev;
{
  if (!editwin_opened)
    return;
  menuwin_enter(menutbl, ev);
}

edit_leave(ev)
     XEvent *ev;
{
  int i;
  
  if (!editwin_opened)
    return 0;
  if (menuwin_leave(menutbl, ev) >= 0)
    return 1;
  return 0;
}

/* End of file */
