/* $Header: /ext/yav/catty/xpx/RCS/file.c,v 1.18 1994/02/02 07:15:43 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"

#define BDW 1
#define MAXFILE 1024
#define MAXFILENAME 64

#define PATHX 8
#define PATHY 8
#define PATHW 300
#define PATHH 18

#define NAMEX 160
#define NAMEY 100
#define NAMEW 140
#define NAMEH 20

#define DIRSLOTMAX 12
#define DIRSLOTX 8
#define DIRSLOTY 64
#define DIRSLOTW 128
#define DIRSLOTH 18

#define SCROLLX DIRSLOTX
#define SCROLLY (DIRSLOTY-(SCROLLH+BDW))
#define SCROLLW DIRSLOTW
#define SCROLLH 20

#define FILEMENUX 196
#define FILEMENUY 160
#define FILEMENUW 64
#define FILEMENUH 20

#define COLNMX PALSETX
#define COLNMY (PALSETY-24)
#define COLNMW 112
#define COLNMH 16
#define IMGNMX COLNMX+COLNMW+8
#define IMGNMY COLNMY
#define IMGNMW COLNMW
#define IMGNMH COLNMH

static Window filewin;
static int filewin_opened = 0;
static int select_slot;
static char filenamebuf[MAXFILE*MAXFILENAME];
static char current_path[MAXPATHLEN];
static char current_path_buf[MAXPATHLEN];
static char current_name[MAXFILENAME];
static char noname[] = "*Noname*";
static int nametop;
static int scroll_dir = 0;
static Window imgfilewin, colfilewin;

#define FILE_PATH	0
#define FILE_NAME	1
#define FILE_OPEN	2
#define FILE_SAVE	3
#define FILE_DUMMY1	4
#define FILE_REMOVE	5
#define FILE_DUMMY2	6
#define FILE_CANCEL	7

char *filemenu_pathstr()
{
  return current_path;
}

char *filemenu_namestr()
{
  return current_name;
}

static MENU filemenu[] = {
  {
    FILE_PATH,
    PATHX, PATHY, PATHW, PATHH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask|KeyPressMask, 0,
    0, 0, BDW,
    filemenu_pathstr, "Path"},
  {
    FILE_NAME,
    NAMEX, NAMEY, NAMEW, NAMEH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask|KeyPressMask, 0,
    0, 0, BDW,
    filemenu_namestr, "Name"},
  {
    FILE_OPEN,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*0, FILEMENUW, FILEMENUH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "Open"},
  {
    FILE_SAVE,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*1, FILEMENUW, FILEMENUH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "Save"},
  {
    FILE_DUMMY1,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*2, FILEMENUW, FILEMENUH,
    0, 0, 0,
    0, 0, BDW,
    NULL, "----"},
  {
    FILE_REMOVE,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*3, FILEMENUW, FILEMENUH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "Remove"},
  {
    FILE_DUMMY2,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*4, FILEMENUW, FILEMENUH,
    0, 0, 0,
    0, 0, BDW,
    NULL, "----"},
  {
    FILE_CANCEL,
    FILEMENUX, FILEMENUY+(FILEMENUH+BDW)*5, FILEMENUW, FILEMENUH,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "Cancel"},
  {
    -1,
    0, 0, 0, 0,
    0, 0, 0,
    0, 0, 0,
    NULL, ""}};

static MENU slotwin[DIRSLOTMAX+2+1];

get_current_path()
{
  get_current_dir(current_path);
  filemenu[FILE_PATH].cursor = strlen(current_path);
}

/*
 * return number of file (0 is error!)
 */
get_filenames()
{
  int i, n;
  FILE *fp;
  int fd[2];
  char buf[MAXPATHLEN+1];

  i = select_slot - nametop;
  if (0 <= i && i < DIRSLOTMAX)
    slotwin[i].mode &= ~MenuModeMaskOn;
  select_slot = nametop = 0;
  slotwin[0].mode |= MenuModeMaskOn;
  n = get_dir(current_path, filenamebuf, MAXFILENAME, MAXFILE);
  for (i = n; i < MAXFILE; i++)
    filenamebuf[i*MAXFILENAME] = '\0';
  XClearArea(dsp, filemenu[FILE_PATH].win, 0, 0, 0, 0, True);
  for (i = 0; i < DIRSLOTMAX; i++)
    XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
  return n;
}

char *slotwin_strfunc(p)
     MENU *p;
{
  return filenamebuf+(nametop+p->n)*MAXFILENAME;
}

create_file_window(x, y)
{
  int i;
  XSizeHints hint;
  char buf[256];

  if (filewin_opened) {
    XMapWindow(dsp, filewin);
    return 1;
  }
  hint.x = x; hint.y = y;
  hint.width = 320; hint.height = 320;
  hint.flags = PPosition|PSize;
  filewin = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
				hint.width, hint.height, BDW, fg, bg);
  sprintf(buf, "%s file", myname);
  XSetStandardProperties(dsp, filewin, buf, buf, None, NULL, NULL, &hint);
  XMapRaised(dsp, filewin);

  menuwin_create(filemenu, filewin, 0, 0);

  x = DIRSLOTX;
  y = DIRSLOTY;
  for (i = 0; i < DIRSLOTMAX; i++) {
    slotwin[i].n = i;
    slotwin[i].x = x;
    slotwin[i].y = y;
    slotwin[i].w = DIRSLOTW;
    slotwin[i].h = DIRSLOTH;
    slotwin[i].mode = MenuModeMaskLeft;
    slotwin[i].bdw = 0;
    slotwin[i].im = ButtonPressMask|ExposureMask|
      EnterWindowMask|LeaveWindowMask;
    slotwin[i].strfunc = slotwin_strfunc;
    y += DIRSLOTH;
  }
  x = SCROLLX;
  y = SCROLLY;
  while (i < DIRSLOTMAX+2) {
    slotwin[i].n = i;
    slotwin[i].x = x;
    slotwin[i].y = y;
    slotwin[i].w = SCROLLW;
    slotwin[i].h = SCROLLH;
    slotwin[i].mode = 0;
    slotwin[i].bdw = BDW;
    slotwin[i].im = ButtonPressMask|ExposureMask|
      EnterWindowMask|LeaveWindowMask|ButtonReleaseMask|OwnerGrabButtonMask;
    slotwin[i].strfunc = NULL;
    strcpy(slotwin[i].str, i == DIRSLOTMAX?"^":"v");
    y += SCROLLH+BDW+DIRSLOTH*DIRSLOTMAX;
    i++;
  }
  slotwin[i].n = -1;
  menuwin_create(slotwin, filewin, 0, 0);

  get_current_path();
  get_filenames();
  filewin_opened = 1;
  return 0;
}

unmap_file_window()
{
  XUnmapWindow(dsp, filewin);
}

close_file_window()
{
  int i;

  if (!filewin_opened)
    return;
  filewin_opened = 0;
  menuwin_destroy(slotwin);
  menuwin_destroy(filemenu);
  XDestroyWindow(dsp, filewin);
}

set_pathcursor()
{
  filemenu[FILE_PATH].cursor = strlen(current_path);
}

change_parent_dir()
{
  char *p;
  extern char *strrchr();
  
  p = strrchr(current_path, '/');
  if (p != NULL) {
    if (p == current_path)
      p++;
    *p = '\0';
  }
  set_pathcursor();
}

void set_imgfilename(name)
     char *name;
{
  strcpy(imgfilename, name);
  XClearArea(dsp, imgfilewin, 0, 0, 0, 0, True);
  update_infowin();
}

void set_colfilename(name)
     char *name;
{
  strcpy(colfilename, name);
  XClearArea(dsp, colfilewin, 0, 0, 0, 0, True);
  update_infowin();
}

create_filename_window()
{
  imgfilewin = XCreateSimpleWindow(dsp, win, IMGNMX, IMGNMY, IMGNMW, IMGNMH,
				   BDW, fg, bg);
  XSelectInput(dsp, imgfilewin, ExposureMask);
  XMapRaised(dsp, imgfilewin);
  colfilewin = XCreateSimpleWindow(dsp, win, COLNMX, COLNMY, COLNMW, COLNMH,
				   BDW, fg, bg);
  XSelectInput(dsp, colfilewin, ExposureMask);
  XMapRaised(dsp, colfilewin);
}

close_filename_window()
{
  XDestroyWindow(dsp, colfilewin);
  XDestroyWindow(dsp, imgfilewin);
}

filename_expose(ev)
     XEvent *ev;
{
  extern char *filname();

  if (ev->xexpose.count)
    return;
  if (ev->xexpose.window == imgfilewin) {
    menuwin_center(imgfilewin, filename(imgfilename), 0);
    return;
  }
  if (ev->xexpose.window == colfilewin) {
    menuwin_center(colfilewin, filename(colfilename), 0);
    return;
  }
}

read_cbb_file(name)
     char *name;
{
  FILE *fp;
  int r;
  unsigned char cbb_data[32*256];
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "cbb file ``%s'' open error!\n", name);
    return 1;
  }
  set_imgfilename(name);
  init_image();
  r = fread(cbb_data, sizeof(cbb_data), 1, fp);
  fclose(fp);
  if (!r) {
    set_imgfilename("");
    fprintf(stderr, "cbb file ``%s'' open error!\n", name);
    return 1;
  }
  imgfilew = 128;
  imgfileh = 128;
  gen_image_cbb(cbb_data);
  image_update_all();
  img_expose(0, 0, imgwinwidth, imgwinheight);
  return 0;
}

read_ccb_file(name)
     char *name;
{
  FILE *fp;
  int r;
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "ccb file ``%s'' open error!\n", name);
    return 1;
  }
  set_colfilename(name);
  r = read_ccb(fp, color_buf);
  fclose(fp);
  if (r == 0) {
    set_colfilename("");
    fprintf(stderr, "ccb file ``%s'' read error!\n", name);
    return 1;
  }
  free_color();
  color_update_all();
  return 0;
}

read_mag_file(name)
     char *name;
{
  FILE *fp;
  int r, fd[2];
  
  pipe(fd);
  if (fork() == 0) {
    close(1);
    dup(fd[1]);
    close(fd[1]);
    close(fd[0]);
    execlp("magtoppm", "magtoppm", name, NULL);
    exit(0);
  }
  fp = fdopen(fd[0], "r");
  close(fd[1]);
  set_colfilename(name);
  set_imgfilename(name);
  r = read_pnm(fp);
  fclose(fp);
  if (r) {
    set_colfilename("");
    set_imgfilename("");
    fprintf(stderr, "mag file ``%s'' read error!\n", name);
  }
  return r;
}

read_pnm_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "pnm file ``%s'' open error!\n", name);
    return 1;
  }
  set_colfilename(name);
  set_imgfilename(name);
  r = read_pnm(fp);
  fclose(fp);
  if (r) {
    set_colfilename("");
    set_imgfilename("");
    fprintf(stderr, "pnm file ``%s'' read error!\n", name);
  }
  return r;
}

read_kcf_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "kcf file ``%s'' open error!\n", name);
    return 1;
  }
  set_colfilename(name);
  r = read_kcf(fp);
  fclose(fp);
  if (r) {
    set_colfilename("");
    fprintf(stderr, "kcf file ``%s'' read error!\n", name);
    return 1;
  }
  return 0;
}

read_cel_file(name)
     char *name;
{
  FILE *fp;
  int r;
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "cel file ``%s'' open error!\n", name);
    return 1;
  }
  set_imgfilename(name);
  r = read_cel(fp);
  fclose(fp);
  if (r) {
    set_imgfilename("");
    fprintf(stderr, "cel file ``%s'' read error!\n", name);
    return 1;
  }
  return 0;
}

read_xbm_file(name)
     char *name;
{
  FILE *fp;
  int r;
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "xbm file ``%s'' open error!\n", name);
    return 1;
  }
  set_imgfilename(name);
  r = read_xbm(fp);
  fclose(fp);
  if (r) {
    set_imgfilename("");
    fprintf(stderr, "xbm file ``%s'' read error!\n", name);
    return 1;
  }
  return 0;
}


write_kcf_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "kcf file ``%s'' write open error!\n", name);
    return 1;
  }
  r = write_kcf(fp, 1);
  fclose(fp);
  if (r) {
    fprintf(stderr, "kcf file ``%s'' write error!\n", name);
    return 1;
  }
  return 0;
}


write_cel_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "cel file ``%s'' write open error!\n", name);
    return 1;
  }
  r = write_cel(fp, 1);
  fclose(fp);
  if (r) {
    fprintf(stderr, "cel file ``%s'' write error!\n", name);
    return 1;
  }
  return 0;
}

write_xbm_file(name)
     char *name;
{
  int r;
  FILE *fp;
  char buf[MAXPATHLEN], *p;
  extern char *basefilename();
  
  strcpy(buf, name);
  p = basefilename(buf);
  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "xbm file ``%s'' write open error!\n", name);
    return 1;
  }
  r = write_xbm(fp, p);
  fclose(fp);
  if (r) {
    fprintf(stderr, "xbm file ``%s'' write error!\n", name);
    return 1;
  }
  return 0;
}


struct {
  char *suffix;
  int (*readfunc)();
  int (*writefunc)();
} file_type_tbl[] = {
  {".ccb", read_ccb_file, NULL},
  {".cbb", read_cbb_file, NULL},
  {".ppm", read_pnm_file, NULL},
  {".mag", read_mag_file, NULL},
  {".kcf", read_kcf_file, write_kcf_file},
  {".cel", read_cel_file, write_cel_file},
  {".xbm", read_xbm_file, write_xbm_file},
  {NULL, NULL, NULL}
};


set_current_name(name)
     char *name;
{
  strcpy(current_name, name);
  filemenu[FILE_NAME].cursor = strlen(current_name);
  XClearArea(dsp, filemenu[FILE_NAME].win, 0, 0, 0, 0, True);
}

reset_current_name()
{
  set_current_name("");
}

get_filetype(name)
     char *name;
{
  int i, n;
  
  for (n = 0; file_type_tbl[n].suffix != NULL; n++) {
    i = strlen(name) - strlen(file_type_tbl[n].suffix);
    if (i >= 0 && strcmp(name+i, file_type_tbl[n].suffix) == 0) {
      return n;
    }
  }
  return -1;
}

read_data_file(name)
     char *name;
{
  int i;

  i = get_filetype(name);
  if (i < 0)
    return 1;
  if (file_type_tbl[i].readfunc == NULL)
    return 1;
  return file_type_tbl[i].readfunc(name);
}

file_open()
{
  int i;
  char buf[MAXPATHLEN];

  i = strlen(current_name);
  if (i && current_name[i-1] == '/') {
    current_name[i-1] = '\0';
    if (strcmp(current_name, "..") == 0)
      change_parent_dir();
    else if (strcmp(current_name, ".")) {
      i = strlen(current_path);
      if (i && current_path[i-1] != '/')
	strcat(current_path, "/");
      strcat(current_path, current_name);
      set_pathcursor();
    }
    get_filenames();
    reset_current_name();
  }
  if (current_name[0] == '\0')
    return;
  sprintf(buf, "%s/%s", current_path, current_name);
  return read_data_file(buf);
}

file_save()
{
  int i;
  char buf[MAXPATHLEN];

  i = strlen(current_name);
  if (!i || current_name[i-1] == '/')
    return 1;
  sprintf(buf, "%s/%s", current_path, current_name);
  i = get_filetype(buf);
  if (i < 0)
    return 1;
  if (file_type_tbl[i].writefunc == NULL)
    return 1;
  return file_type_tbl[i].writefunc(buf);
}

file_remove()
{
  int i;
  char buf[MAXPATHLEN];

  i = strlen(current_name);
  if (!i || current_name[i-1] == '/')
    return;
  sprintf(buf, "%s/%s", current_path, current_name);
  if (unlink(buf) == 0) {
    for (i = select_slot; i < MAXFILE-1; i++) {
      if (filenamebuf[i*MAXFILENAME] == '\0')
	break;
      strcpy(filenamebuf+i*MAXFILENAME, filenamebuf+(i+1)*MAXFILENAME);
    }
    set_current_name(filenamebuf+(select_slot)*MAXFILENAME);
    for (i = select_slot - nametop; 0 <= i && i < DIRSLOTMAX; i++)
      XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
  }
}

file_expose(ev)
     XEvent *ev;
{
  int i;

  if (!filewin_opened)
    return;
  if (ev->xexpose.count != 0)
    return;
  if (menuwin_expose(filemenu, ev) >= 0)
    return;
  menuwin_expose(slotwin, ev);
}

scroll_down()
{
  int i;
  
  if (nametop == 0)
    return;
  --nametop;
  for (i = DIRSLOTMAX-1; i > 0; --i) {
    if (slotwin[i-1].mode & MenuModeMaskOn)
      slotwin[i].mode |= MenuModeMaskOn;
    else
      slotwin[i].mode &= ~MenuModeMaskOn;
    XCopyArea(dsp, slotwin[i-1].win, slotwin[i].win, gc0, 0, 0,
	      DIRSLOTW, DIRSLOTH, 0, 0);
  }
  if (nametop == select_slot)
    slotwin[0].mode |= MenuModeMaskOn;
  else
    slotwin[0].mode &= ~MenuModeMaskOn;
  XClearArea(dsp, slotwin[0].win, 0, 0, 0, 0, True);
}

scroll_up()
{
  int i;
  
  if (filenamebuf[(nametop+DIRSLOTMAX-1)*MAXFILENAME] == '\0')
    return;
  ++nametop;
  for (i = 0; i < DIRSLOTMAX-1; i++) {
    if (slotwin[i+1].mode & MenuModeMaskOn)
      slotwin[i].mode |= MenuModeMaskOn;
    else
      slotwin[i].mode &= ~MenuModeMaskOn;
    XCopyArea(dsp, slotwin[i+1].win, slotwin[i].win, gc0, 0, 0,
	      DIRSLOTW, DIRSLOTH, 0, 0);
  }
  if (nametop+DIRSLOTMAX-1 == select_slot)
    slotwin[DIRSLOTMAX-1].mode |= MenuModeMaskOn;
  else
    slotwin[DIRSLOTMAX-1].mode &= ~MenuModeMaskOn;
  XClearArea(dsp, slotwin[DIRSLOTMAX-1].win, 0, 0, 0, 0, True);
}

file_button(ev)
     XEvent *ev;
{
  int i, oldsel;

  if (!filewin_opened)
    return;
  i = menuwin_press(slotwin, ev);
  if (i >= 0) {
    switch(i) {
    case DIRSLOTMAX:
      scroll_dir = -1;
      XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
      break;
    case DIRSLOTMAX+1:
      scroll_dir = 1;
      XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
      break;
    default:
      if (filenamebuf[(nametop+i)*MAXFILENAME] == '\0') {
	slotwin[i].mode &= ~MenuModeMaskOn;
	XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
      } else {
	oldsel = select_slot;
	select_slot = nametop+i;
	set_current_name(filenamebuf+select_slot*MAXFILENAME);
	i = oldsel - nametop;
	if (0 <= i && i < DIRSLOTMAX) {
	  slotwin[i].mode &= ~MenuModeMaskOn;
	  XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
	}
	i = select_slot - nametop;
	slotwin[i].mode |= MenuModeMaskOn;
	XClearArea(dsp, slotwin[i].win, 0, 0, 0, 0, True);
	if (double_clicked)
	  file_open();
      }
      break;
    }
    return;
  }
  i = menuwin_press(filemenu, ev);
  if (i >= 0) {
    switch(i) {
    case FILE_OPEN:
      file_open();
      break;
    case FILE_SAVE:
      file_save();
      break;
    case FILE_DUMMY1:
      break;
    case FILE_REMOVE:
      file_remove();
      break;
    case FILE_DUMMY2:
      break;
    case FILE_CANCEL:
      menu_stat_change(1, 0);
      break;
    }
    filemenu[i].mode &= ~MenuModeMaskOn;	/* every off */
    XClearArea(dsp, filemenu[i].win, 0, 0, 0, 0, True);
    return;
  }
}

file_release(ev)
     XEvent *ev;
{
  if (ev->xbutton.window == slotwin[DIRSLOTMAX].win) {
    scroll_dir = 0;
    slotwin[DIRSLOTMAX].mode &= ~MenuModeMaskOn;
    XClearArea(dsp, slotwin[DIRSLOTMAX].win, 0, 0, 0, 0, True);
  } else if (ev->xbutton.window == slotwin[DIRSLOTMAX+1].win) {
    scroll_dir = 0;
    slotwin[DIRSLOTMAX+1].mode &= ~MenuModeMaskOn;
    XClearArea(dsp, slotwin[DIRSLOTMAX+1].win, 0, 0, 0, 0, True);
  }
}

file_enter(ev)
     XEvent *ev;
{
  int i;

  if (!filewin_opened)
    return;
  i = menuwin_enter(slotwin, ev);
  if (i >= 0) {
    return;
  }
  i = menuwin_enter(filemenu, ev);
  if (i >= 0) {
    switch(i) {
    case FILE_PATH:
      strcpy(current_path_buf, current_path);
      break;
    }
    return;
  }
}

file_leave(ev)
     XEvent *ev;
{
  int i;
  
  i = menuwin_leave(slotwin, ev);
  if (i >= 0)
    return 1;
  i = menuwin_leave(filemenu, ev);
  if (i >= 0) {
    switch(i) {
    case FILE_PATH:
      strcpy(current_path, current_path_buf);
      set_pathcursor();
      break;
    }
    return 1;
  }
  return 0;
}

file_keypress(ev)
     XEvent *ev;
{
  int i, r;
  
  i = menuwin_keypress(filemenu, ev, &r);
  if (i >= 0) {
    switch(i) {
    case FILE_PATH:
      switch(r) {
      case 3:
	strcpy(current_path, current_path_buf);
	set_pathcursor();
      case 2:
	if (get_filenames() == 0) {
	  strcpy(current_path, current_path_buf);
	  get_filenames();
	  set_pathcursor();
	}
	strcpy(current_path_buf, current_path);
      }
    }
    if (r)
      XClearArea(dsp, filemenu[i].win, 0, 0, 0, 0, True);
    return 1;
  }
  return 0;
}

file_background()
{
  switch (scroll_dir) {
  case 1:
    scroll_up();
    break;
  case -1:
    scroll_down();
    break;
  }
}

/* End of file */
