/* $Header: /ext/yav/catty/xpx/RCS/info.c,v 1.4 1994/02/02 07:14:54 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#include "infodef.h"

#define BDW 1

#define INFOWIDTH 480
#define INFOHEIGHT 180
#define INFOLINESPACE 24

#define VIEWX 8
#define VIEWY (INFOHEIGHT-1)

static MENU menutbl[] = {
  {
    INFOMENU_IMGWIDTH,
    60, (INFOLINESPACE*2+6), 40, 20,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask|KeyPressMask, 0,
    0, 0, BDW,
    NULL, "?"},
  {
    INFOMENU_IMGHEIGHT,
    120, (INFOLINESPACE*2+6), 40, 20,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask|KeyPressMask, 0,
    0, 0, BDW,
    NULL, "?"},
  {
    INFOMENU_GRIDCOLOR,
    4, (INFOLINESPACE*5+6), 100, 20,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "grid color"},
  {
    INFOMENU_SCROLLSTOP,
    128, (INFOLINESPACE*5+6), 100, 20,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, 0,
    0, 0, BDW,
    NULL, "scroll stop"},
  {
    INFOMENU_OK,
    (INFOWIDTH-50-BDW*2-4), (INFOHEIGHT-20-BDW*2-4), 50, 20,
    0, EnterWindowMask|LeaveWindowMask|ButtonPressMask, SouthEastGravity,
    0, 0, BDW,
    NULL, "OK"},
  {
    -1,
    0, 0, 0, 0,
    0, 0, 0,
    0, 0, 0,
    NULL, ""}
};
static Window infowin;
static int infowin_opened = 0;

info_imgsize_set()
{
  char *p;
  
  if (!infowin_opened)
    return;
  p = menutbl[INFOMENU_IMGWIDTH].str;
  sprintf(p, "%d", imgfilew);
  menutbl[INFOMENU_IMGWIDTH].cursor = strlen(p);
  p = menutbl[INFOMENU_IMGHEIGHT].str;
  sprintf(p, "%d", imgfileh);
  menutbl[INFOMENU_IMGHEIGHT].cursor = strlen(p);
  XClearArea(dsp, menutbl[INFOMENU_IMGWIDTH].win, 0, 0, 0, 0, True);
  XClearArea(dsp, menutbl[INFOMENU_IMGHEIGHT].win, 0, 0, 0, 0, True);
}

create_info_window(x, y)
{
  int i;
  XSizeHints hint;
  char *p;
  char buf[256];

  if (infowin_opened) {
    XMapWindow(dsp, infowin);
    return 1;
  }
  hint.x = x; hint.y = y;
  hint.width = INFOWIDTH; hint.height = INFOHEIGHT;
  hint.flags = PPosition|PSize;
  infowin = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
				hint.width, hint.height, BDW, fg, bg);
  sprintf(buf, "%s info", myname);
  XSetStandardProperties(dsp, infowin, buf, buf, None, NULL, NULL, &hint);
  XSelectInput(dsp, infowin, ExposureMask);
  XMapRaised(dsp, infowin);

  viewwin = XCreateSimpleWindow(dsp, infowin, VIEWX, VIEWY,
				imgmaxw, imgmaxh, BDW, fg, bg);
  XSelectInput(dsp, viewwin, ExposureMask|
	       ButtonPressMask|EnterWindowMask|LeaveWindowMask);
  XMapRaised(dsp, viewwin);

  menuwin_create(menutbl, infowin, 0, 0);
  menutbl[INFOMENU_SCROLLSTOP].mode &= ~MenuModeMaskOn;
  menutbl[INFOMENU_SCROLLSTOP].mode |= scroll_stop & MenuModeMaskOn;
  infowin_opened = 1;
  info_imgsize_set();
  return 0;
}

unmap_info_window()
{
  menutbl[INFOMENU_OK].mode &= ~MenuModeMaskOn;
  XUnmapWindow(dsp, infowin);
}

close_info_window()
{
  int i;

  if (!infowin_opened)
    return;
  infowin_opened = 0;
  menuwin_destroy(menutbl);
  XDestroyWindow(dsp, viewwin);
  XDestroyWindow(dsp, infowin);
}


redraw_info()
{
  int i, x, y;
  char *p;
  char buf[MAXPATHLEN+32];
  static char *infotbl[] = {
    "* %s %s *",
    "image file : %s",
    "size :",
    "color file : %s", 
    "color : %d (private %d common %d)",
    NULL};

  x = 4;
  y = 0;
  for (i = 0; (p = infotbl[i]) != NULL; i++) {
    y += INFOLINESPACE;
    switch(i) {
    case 0:
      sprintf(buf, p, myname, rcsrev);
      break;
    case 1:
      sprintf(buf, p, imgfilename);
      break;
    case 2:
      XDrawString(dsp, infowin, gc0,
		  (menutbl[INFOMENU_IMGWIDTH].x+menutbl[INFOMENU_IMGWIDTH].w+
		   menutbl[INFOMENU_IMGHEIGHT].x-XTextWidth(fs0, "x", 1)) / 2,
		  y, "x", 1);
      strcpy(buf, p);
      break;
    case 3:
      sprintf(buf, p, colfilename);
      break;
    case 4:
      sprintf(buf, p, alloced_color,
	      alloced_color-alloced_common_color,
	      alloced_common_color);
      break;
    default:
      strcpy(buf, p);
      break;
    }
    XDrawString(dsp, infowin, gc0, x, y, buf, strlen(buf));
  }
}

update_infowin()
{
  if (infowin_opened)
    XClearArea(dsp, infowin, 0, 0, 0, 0, True);
}

alloc_vimg()
{
  vimg = XCreateImage(dsp, DefaultVisual(dsp, scr), DefaultDepth(dsp, scr),
		      ZPixmap, 0, None, imgfilew, imgfileh, 32, 0);
  if (vimg == NULL) {
    fprintf(stderr, "view image alloc error!\n");
    return 1;
  }
  vimg->data = (char *)malloc(vimg->bytes_per_line * vimg->height);
  if (vimg->data == NULL) {
    fprintf(stderr, "view image alloc error!\n");
    return 1;
  }
  return 0;
}

free_vimg()
{
  if (vimg != NULL) {
    if (vimg->data != NULL)
      free(vimg->data);
    XFree((char *)vimg);
    vimg = NULL;
  }
}

make_view_image()
{
  if (view_sw & 1) {
    free_vimg();
    if (vimg == NULL) {
      if (alloc_vimg())
	return 1;
      gen_image_subv(0, 0, imgfilew, imgfileh);
    }
  }
}

redraw_view(x, y, w, h)
{
  if (view_sw & 1) {
    if (vimg == NULL)
      make_view_image();
    XPutImage(dsp, viewwin, gc0, vimg, x, y, x, y, w, h);
  } else {
    menuwin_center(viewwin, "view off", 0);
  }
}

info_expose(ev)
     XEvent *ev;
{
  int i;
  char buf[256];

  if (!infowin_opened)
    return;
  if (ev->xexpose.window == viewwin) {
    redraw_view(ev->xexpose.x, ev->xexpose.y,
		ev->xexpose.width, ev->xexpose.height);
    return;
  }
  if (ev->xexpose.count != 0)
    return;
  if (ev->xexpose.window == infowin) {
    redraw_info();
    return;
  }
  menuwin_expose(menutbl, ev);
}


info_press(ev)
     XEvent *ev;
{
  int i, oldsel;

  if (!infowin_opened)
    return;
  if (ev->xbutton.window == viewwin) {
    view_sw ^= 1;
    XClearArea(dsp, viewwin, 0, 0, 0, 0, True);
    return;
  }
  i = menuwin_press(menutbl, ev);
  if (i >= 0) {
    switch(i) {
    case INFOMENU_GRIDCOLOR:
      XSetForeground(dsp, gridgc, (menutbl[i].mode & MenuModeMaskOn)?fg:bg);
      img_expose(0, 0, imgwinwidth, imgwinheight);
      break;
    case INFOMENU_SCROLLSTOP:
      scroll_stop = menutbl[i].mode & MenuModeMaskOn;
      break;
    case INFOMENU_OK:
      menu_stat_change(2, 0);
      break;
    }
    XClearArea(dsp, menutbl[i].win, 0, 0, 0, 0, True);
  }
}

info_enter(ev)
     XEvent *ev;
{
  if (!infowin_opened)
    return;
  if (ev->xcrossing.window == viewwin) {
    if (!(view_sw & 1))
      menuwin_highlight(viewwin, 0);
    return;
  }
  menuwin_enter(menutbl, ev);
}

info_leave(ev)
     XEvent *ev;
{
  int i;
  
  if (!infowin_opened)
    return 0;
  if (ev->xcrossing.window == viewwin) {
    if (view_sw & 1)
      return 1;
  }
  if (menuwin_leave(menutbl, ev) >= 0)
    return 1;
  return 0;
}

info_keypress(ev)
     XEvent *ev;
{
  int i, r;
  int *ip, x;
  
  if (!infowin_opened)
    return 0;
  i = menuwin_keypress(menutbl, ev, &r);
  if (i >= 0) {
    switch (i) {
    case INFOMENU_IMGWIDTH:
      ip = &imgfilew;
      break;
    case INFOMENU_IMGHEIGHT:
      ip = &imgfileh;
      break;
    default:
      ip = NULL;
    }
    if (r == 2 && ip != NULL) {
      x = strtol(menutbl[i].str, NULL, 0);
      switch(i) {
      case INFOMENU_IMGWIDTH:
	set_image_size(x, -1);
	break;
      case INFOMENU_IMGHEIGHT:
	set_image_size(-1, x);
	break;
      }
      image_update_all();
    }
    if (r)
      XClearArea(dsp, menutbl[i].win, 0, 0, 0, 0, True);
    return 1;
  }
  return 0;
}

/* End of file */
